#include "dict.h"
#include "defs.h"
#include <stdlib.h>
#include <stdint.h>

void init_movelist(movelist *l)
{
  l->v = NULL;
  l->size = l->maxsize = 0;
}

void free_movelist(movelist *l)
{
  if (l->maxsize) free(l->v);
}

void add_movelist(movelist *l, move *m)
{
  if (l->size == l->maxsize) {
    l->maxsize += 1024;
    l->v = realloc(l->v, l->maxsize * sizeof(move)); if (l->v==NULL) abort();
  }
  l->v[l->size] = *m;
  l->size++;
}

int is_letter(game *g, int i, int j)
{
  return i >= 0 && i < 15 && j >= 0 && j < 15 && g->board[i][j] != 0;
}

static void check_word(game *g, letterstack *s, movelist *ml,
                int starti, int startj, int inci, int incj)
{
  /* to be ok a word must include an already positioned letter or touch one
   * or be put on the center position (can only happen for first word)
   * note that when check_word is called we already checked that putting
   * the letters on the board give existing word (or words)
   */
  move m;
  int i, j, l, ok = 0;
  /* word must contain at least 1 new letter */
  for (l=0, i=starti, j=startj; l<s->stacklen; l++, i+=inci, j+=incj)
    if (!is_letter(g, i, j)) { ok = 1; break; }
  if (!ok) return;
  ok = 0;
  for (l=0, i=starti, j=startj; l<s->stacklen; l++, i+=inci, j+=incj)
    if (is_letter(g, i, j) ||             /* word contain letter on board */
        is_letter(g, i-incj, j-inci) ||   /* word touch letter on board */
        is_letter(g, i+incj, j+inci) ||   /* word touch letter on board */
        (i == 7 && j == 7))            /* word is at center pos (first word) */
    { ok = 1; break; }
  if (!ok) return;
  for (l = 0; l < 7; l++) m.next_hand[l] = g->hand[l];
  for (l = 0; l < s->stacklen; l++) {
    m.word[l] = s->stack[l].l->l;
    if (s->stack[l].hl == 0) m.word[l] += 26;
    if (s->stack[l].new) m.word[l] = -m.word[l];
  }
  m.wordlen = s->stacklen;
  m.row = starti;
  m.col = startj;
  m.vert = inci ? 1 : 0;
  m.score = score(g, m.word, m.wordlen, starti, startj, inci, incj);
  add_movelist(ml, &m);
}

static void push(letterstack *s, game *g, dict_letter *n, int hl, int new)
{
  s->stack[s->stacklen].l = n;
  s->stack[s->stacklen].hl = hl;
  s->stack[s->stacklen].new = new;
  s->stacklen++;
}

static void pop(letterstack *s)
{
  s->stacklen--;
}

static int other_word_ok(dictionary *d, game *g, dict_letter *n, int i, int j,
                         int inci, int incj)
{
  int si = i, sj = j;
  int word[16];
  int w;
  int ii = 1 - inci;
  int jj = 1 - incj;
  int starti = i, startj = j;
  int endi, endj;
  g->board[si][sj] = n->l;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && g->board[starti][startj] != 0)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && g->board[endi][endj] != 0)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj) {
    int letter = abs(g->board[i][j]); if (letter > 26) letter -= 26;
    word[w] = letter;
  }
  word[w] = 0;
  g->board[si][sj] = 0;
  return w == 1 || find_word(d, word);
}

static void go(dictionary *dict, game *g, movelist *l, letterstack *s,
               dict_letter *cur,
               int starti, int startj, int i, int j, int inci, int incj)
{
  int h;
  int32_t letter_done = 0;  /* bit field, 27 bits, 1 bit per letter */
  int32_t joker_done = 0;  /* bit field, 27 bits, 1 bit per letter for joker */

  if (cur == NULL) return;
  if (i == 15 || j == 15) return;

  if (g->board[i][j] != 0) {
    /* must match board[i][j] */
    while (cur) {
      int letter = abs(g->board[i][j]); if (letter > 26) letter -= 26;
      if (cur->l == letter) {
        push(s, g, cur, cur->l, 0);
        /* if n->w we are in an end letter of a word in the dictionary
         * so if on the board there is nothing after current letter,
         * we check that the word is ok here at (i,j)
         */
        if (cur->w && (i+inci==15||j+incj==15||g->board[i+inci][j+incj]==0))
          check_word(g, s, l, starti, startj, inci, incj);
        go(dict, g, l, s, cur->child,
           starti, startj, i + inci, j + incj, inci, incj);
        pop(s);
        return;
      }
      cur = cur->next;
    }
    return;
  }

  /* must match one of not-yet-used letters */
  for (h = 0; h < 7; h++) {
    dict_letter *n = cur;
    int hletter;
    if (g->hand[h] == -1) continue;
    hletter = g->hand[h];
    g->hand[h] = -1;
    while (n) {
      /* letter must match or g is joker
       * and we must not have tried that letter yet
       * and (eventual) word in other direction must be ok */
      if ((n->l == hletter || hletter == 0)
          && !(hletter && (letter_done & (1 << hletter)))
          && !(!hletter && (joker_done & (1 << n->l)))
          && other_word_ok(dict, g, n, i, j, inci, incj)) {
        if (hletter) letter_done |= 1 << hletter;
        else         joker_done |= 1 << n->l;
        push(s, g, n, hletter, 1);
        if (n->w && (i+inci==15 || j+incj==15 || g->board[i+inci][j+incj]==0))
          check_word(g, s, l, starti, startj, inci, incj);
        go(dict, g, l, s, n->child,
           starti, startj, i + inci, j + incj, inci, incj);
        pop(s);
      }
      n = n->next;
    }
    g->hand[h] = hletter;
  }
}

void get_moves(game *g, movelist *l, dictionary *dict)
{
  letterstack s;
  int i, j;

  s.stacklen = 0;

  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    /* must be a start of a word (no letter before) */
    if (is_letter(g, i-1, j-0)) continue;
    go(dict, g, l, &s, dict->top, i, j, i, j, 1, 0);
  }
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    /* must be a start of a word (no letter before) */
    if (is_letter(g, i-0, j-1)) continue;
    go(dict, g, l, &s, dict->top, i, j, i, j, 0, 1);
  }

}
